/* -*- Mode: C++; c-default-style: "k&r"; indent-tabs-mode: nil; tab-width: 2; c-basic-offset: 2 -*- */

/* libstaroffice
* Version: MPL 2.0 / LGPLv2+
*
* The contents of this file are subject to the Mozilla Public License Version
* 2.0 (the "License"); you may not use this file except in compliance with
* the License or as specified alternatively below. You may obtain a copy of
* the License at http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* Major Contributor(s):
* Copyright (C) 2002 William Lachance (wrlach@gmail.com)
* Copyright (C) 2002,2004 Marc Maurer (uwog@uwog.net)
* Copyright (C) 2004-2006 Fridrich Strba (fridrich.strba@bluewin.ch)
* Copyright (C) 2006, 2007 Andrew Ziem
* Copyright (C) 2011, 2012 Alonso Laurent (alonso@loria.fr)
*
*
* All Rights Reserved.
*
* For minor contributions see the git repository.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU Lesser General Public License Version 2 or later (the "LGPLv2+"),
* in which case the provisions of the LGPLv2+ are applicable
* instead of those above.
*/

#ifndef STOFF_CELL_STYLE
#  define STOFF_CELL_STYLE

#include <string>
#include <vector>

#include "libstaroffice_internal.hxx"

//! Class to store a cell style
class STOFFCellStyle
{
public:
  /** constructor */
  STOFFCellStyle() : m_propertyList(), m_numberCellSpanned(1,1), m_format(0)
  {
  }
  //! add to the propList
  void addTo(librevenge::RVNGPropertyList &propList) const;

  //! operator<<
  friend std::ostream &operator<<(std::ostream &o, STOFFCellStyle const &cellStyle);
  //! operator==
  bool operator==(STOFFCellStyle const &cellStyle) const;
  //! operator!=
  bool operator!=(STOFFCellStyle const &cellStyle) const
  {
    return !operator==(cellStyle);
  }
  /** the property list */
  librevenge::RVNGPropertyList m_propertyList;
  //! the cell spanned : by default (1,1)
  STOFFVec2i m_numberCellSpanned;
  //! the cell format : 0 by default
  unsigned m_format;
};


#endif
// vim: set filetype=cpp tabstop=2 shiftwidth=2 cindent autoindent smartindent noexpandtab:
